﻿using System;
using System.Net.Http;
using System.Net.Http.Headers;
using System.Text.Json;
using System.Threading;
using System.Threading.Tasks;
using Xamarin.Essentials;

namespace Gallery.Util
{
    public static class NetHelper
    {
        public static bool NetworkAvailable
        {
            get
            {
                try
                {
                    return Connectivity.NetworkAccess == NetworkAccess.Internet
                        || Connectivity.NetworkAccess == NetworkAccess.ConstrainedInternet;
                }
                catch
                {
                    return false;
                }
            }
        }

        public static async Task<(T result, string error)> RequestObject<T>(string url,
            string referer = null,
            HttpContent post = null,
            Action<HttpRequestHeaders> headerHandler = null,
            Func<string, string> contentHandler = null,
            Func<string, T> @return = null)
        {
            var response = await Request(url, headers =>
            {
                if (referer != null)
                {
                    headers.Referrer = new Uri(referer);
                }
                headers.Add("User-Agent", Config.UserAgent);
                headerHandler?.Invoke(headers);
            }, post);
            if (response == null)
            {
                return (default, "response is null");
            }
            if (!response.IsSuccessStatusCode)
            {
                Log.Print($"http failed with code: {(int)response.StatusCode} - {response.StatusCode}");
                return (default, response.StatusCode.ToString());
            }
            string content;
            using (response)
            {
                try
                {
                    content = await response.Content.ReadAsStringAsync();
                    if (contentHandler != null)
                    {
                        content = contentHandler(content);
                    }
                    if (@return != null)
                    {
                        return (@return(content), null);
                    }
                }
                catch (Exception ex)
                {
                    Log.Error("stream.load", $"failed to read stream, error: {ex.Message}");
                    return (default, ex.Message);
                }
            }

            if (content == null)
            {
                content = string.Empty;
            }
            try
            {
                var result = JsonSerializer.Deserialize<T>(content);
                return (result, null);
            }
            catch (Exception ex)
            {
                var memo = content.Length < 20 ? content : content[0..20] + "...";
                Log.Error("content.deserialize", $"failed to parse JSON object, content: {memo}, error: {ex.Message}");
                return (default, content);
            }
        }

        private static async Task<HttpResponseMessage> Request(string url, Action<HttpRequestHeaders> headerHandler, HttpContent post = null)
        {
#if DEBUG
            var method = post == null ? "GET" : "POST";
            Log.Print($"{method}: {url}");
#endif
            var uri = new Uri(url);
            var proxy = Config.Proxy;
            var handler = new HttpClientHandler
            {
                AutomaticDecompression = System.Net.DecompressionMethods.GZip | System.Net.DecompressionMethods.Deflate,
                UseCookies = false
            };
            if (proxy != null)
            {
                handler.Proxy = proxy;
                handler.UseProxy = true;
            }
            var client = new HttpClient(handler, true)
            {
                BaseAddress = new Uri($"{uri.Scheme}://{uri.Host}:{uri.Port}"),
                Timeout = Config.Timeout
            };
            return await TryCount(() =>
            {
                using var request = new HttpRequestMessage(post == null ? HttpMethod.Get : HttpMethod.Post, uri.PathAndQuery);
                var headers = request.Headers;
                headerHandler?.Invoke(headers);
                headers.Add("Accept-Language", Config.AcceptLanguage);
                if (post != null)
                {
                    request.Content = post;
                }
                return client.SendAsync(request, HttpCompletionOption.ResponseHeadersRead);
            });
        }

        private static T TryCount<T>(Func<T> func, int tryCount = 2)
        {
            int tries = 0;
            while (tries < tryCount)
            {
                try
                {
                    return func();
                }
                catch (Exception ex)
                {
                    tries++;
                    Thread.Sleep(1000);
                    Log.Error("try.do", $"tries: {tries}, error: {ex.Message}");
                }
            }
            return default;
        }
    }
}
